#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>

#define DICE (6)
#define MAX_FAIL(dice) (6)
#define MIN_SUCCESS(dice) (10)
#define ROLL(val) (1 + (random() % (val)))
#define ADV_BONUS(dice) (1)
#define THIRST_BONUS(dice) (1)

#define MIN(a, b) (((a) < (b)) ? (a) : (b))
#define MAX(a, b) (((a) > (b)) ? (a) : (b))

_Bool is_do_rerol(int dice, int bonus, int advantage, _Bool is_thirst, int result)
{
	int real_bonus = bonus;

	if (advantage > 0)
		real_bonus += ADV_BONUS(dice);
	else if (advantage < 0)
		real_bonus -= ADV_BONUS(dice);

	if (is_thirst)
		real_bonus += THIRST_BONUS(dice);

	int medium_val = dice + 1;

	_Bool rerol_res = (result <= MAX_FAIL(dice)) || (medium_val + real_bonus >= MIN_SUCCESS(dice) && (result < MIN_SUCCESS(dice)));

	return rerol_res;
}


int roll_internal(int dice, int bonus, int advantage, _Bool is_thirst)
{
	int d11 = ROLL(dice), d12 = ROLL(dice);
	int d1 = (advantage == 0) ? d11
		: ((advantage > 0) ? MAX(d11, d12) : MIN(d11, d12));
	int d2 = (is_thirst) ? ROLL(dice + 2) : ROLL(dice);

	int res = d1 + d2 + bonus;

	return res;
}

int roll(int dice, int bonus, int advantage, _Bool is_thirst, _Bool is_rerol)
{
	int result = roll_internal(dice, bonus, advantage, is_thirst);

	if (is_rerol && is_do_rerol(dice, bonus, advantage, is_thirst, result))
		result = roll_internal(dice, bonus, advantage, is_thirst);

	return result;
}

int roll_x3(int dice, int bonus)
{
	int d1 = ROLL(dice), d2 = ROLL(dice), d3 = ROLL(dice);
	int min = MIN(d1, MIN(d2, d3));

	return d1 + d2 + d3 - min + bonus;
}

int roll_x2x2(int dice, int bonus)
{
	int d1 = ROLL(dice), d2 = ROLL(dice), d3 = ROLL(dice), d4 = ROLL(dice);

	return MAX(d1, d2) + MAX(d3, d4) + bonus;
}

#define MIN_BONUS (-5)
#define MAX_BONUS (5)
#define VARIANTS (MAX_BONUS - MIN_BONUS + 1)
#define COUNT (100000)
#define COUNT_D ((double)COUNT)
#define MAX_VALUE (14)
#define DATA_SIZE (MAX_VALUE + 1 + MAX_BONUS - MIN_BONUS)

typedef int data_arr_t[VARIANTS][DATA_SIZE];
typedef struct {
	long long var[VARIANTS];
	long long fail[VARIANTS], partial[VARIANTS], succ[VARIANTS];
	long long res[VARIANTS];
} aggr_data_t;

void inc_data_arr(data_arr_t arr, int bonus, int res)
{
	assert(bonus - MIN_BONUS >= 0);
	assert(bonus - MIN_BONUS < VARIANTS);
	assert(res - MIN_BONUS >= 0);
	assert(res - MIN_BONUS < DATA_SIZE);
	arr[bonus-MIN_BONUS][res-MIN_BONUS]++;
}

long long get_data_arr(data_arr_t arr, int bonus, int res)
{
	assert(bonus - MIN_BONUS >= 0);
	assert(bonus - MIN_BONUS < VARIANTS);
	assert(res - MIN_BONUS >= 0);
	assert(res - MIN_BONUS < DATA_SIZE); 
	return arr[bonus-MIN_BONUS][res-MIN_BONUS];
}

double get_data_arr_d(data_arr_t arr, int bonus, int res)
{
	return (double)(get_data_arr(arr, bonus, res));
}

#define DIS 0
#define NORM 1
#define ADV 2
#define THIRST_DIS 3
#define THIRST_NORM 4
#define THIRST_ADV 5

#define RE_DIS 6
#define RE_NORM 7
#define RE_ADV 8
#define RE_THIRST_DIS 9
#define RE_THIRST_NORM 10
#define RE_THIRST_ADV 11

#define USE_CASE_COUNT 12

#define GET_ADV(num) (((num) % 3) - 1)
#define GET_THIRST(num) (((num) / 3) % 2)
#define GET_REROL(num) (((num) / 6) % 2)
char *names[USE_CASE_COUNT] = {
		"DIS      ", "NORM     ", "ADV      ", "THIRST_DIS   ", "THIRST_NORM   ", "THIRST_ADV   ",
		"RE_DIS   ", "RE_NORM  ", "RE_ADV   ", "RE_THIRST_DIS", "RE_THIRST_NORM", "RE_THIRST_ADV"};

int main(int argc, char **argv)
{
	bool is_verbose = (argc > 1) && !strcmp(argv[1], "--verbose");

	srand(time(NULL));
	data_arr_t data[USE_CASE_COUNT] = {0};

	int bonus, i, j;
	for (j = 0; j < USE_CASE_COUNT; j++)
		printf("%2d:\t\t%4d  %4d  %4d\n", j, GET_ADV(j), GET_THIRST(j), GET_REROL(j));
	for (bonus = MIN_BONUS; bonus <= MAX_BONUS; bonus++) {
		for (i = 0; i < COUNT; i++) {
			for (j = 0; j < USE_CASE_COUNT; j++) {
				inc_data_arr(data[j], bonus, roll(DICE, bonus, GET_ADV(j), GET_THIRST(j), GET_REROL(j)));
			}
		}
	}

	aggr_data_t summ[USE_CASE_COUNT] = {0};
	for (bonus = MIN_BONUS; bonus <= MAX_BONUS; bonus++) {
		if (is_verbose)
			printf("\nBONUS == %d\n", bonus);
		for (i = MIN_BONUS; i <= MAX_VALUE+MAX_BONUS; i++) {

			for (j = 0; j < USE_CASE_COUNT; j++) {
				summ[j].var[bonus-MIN_BONUS] += get_data_arr(data[j], bonus, i);
			}

			if (i == MAX_FAIL(DICE)) {
				for (j = 0; j < USE_CASE_COUNT; j++) {
					summ[j].fail[bonus-MIN_BONUS] = summ[j].var[bonus-MIN_BONUS];
				}
			} else if (i == MIN_SUCCESS(DICE)-1) {
				for (j = 0; j < USE_CASE_COUNT; j++) {
					summ[j].partial[bonus-MIN_BONUS] = summ[j].var[bonus-MIN_BONUS] - summ[j].fail[bonus-MIN_BONUS];
				}
			} else if (i == MAX_VALUE) {
				for (j = 0; j < USE_CASE_COUNT; j++) {
					summ[j].succ[bonus-MIN_BONUS] = COUNT - summ[j].partial[bonus-MIN_BONUS] - summ[j].fail[bonus-MIN_BONUS];
					summ[j].res[bonus-MIN_BONUS] = summ[j].succ[bonus-MIN_BONUS] - summ[j].fail[bonus-MIN_BONUS];
				}
			}

			if (is_verbose) {
				printf("%2d: ", i);
				for (j = 0; j < USE_CASE_COUNT; j++) {
					printf("%s=%5.2f(sum=%5.2f); ", names[j],
							36.0*get_data_arr_d(data[j], bonus, i)/COUNT_D,
							36.0*summ[j].var[bonus-MIN_BONUS]/COUNT_D);
				}
				printf("\n");
			}
		}
	}

	printf("\n\n");
	for (bonus = MIN_BONUS+1; bonus < MAX_BONUS; bonus++) {
		printf("\nBONUS == %d\n", bonus);
		printf("\t%s(-1 / +1)\t", names[NORM]);
		for (j = 0; j < USE_CASE_COUNT; j++) {
			if (j != NORM)
				printf("%s\t", names[j]);
		}
		printf("\n");

		printf("6-:\t%5.2f(%5.2f | %5.2f)\t",
				36.0*summ[NORM].fail[bonus-MIN_BONUS]/COUNT_D,
					36.0*(summ[NORM].fail[bonus-MIN_BONUS-1] - summ[NORM].fail[bonus-MIN_BONUS])/COUNT_D,
					36.0*(summ[NORM].fail[bonus-MIN_BONUS+1] - summ[NORM].fail[bonus-MIN_BONUS])/COUNT_D);
		for (j = 0; j < USE_CASE_COUNT; j++) {
			if (j != NORM)
				printf("%5.2f(%5.2f)\t",
					36.0*summ[j].fail[bonus-MIN_BONUS]/COUNT_D,
					36.0*(summ[j].fail[bonus-MIN_BONUS] - summ[NORM].fail[bonus-MIN_BONUS])/COUNT_D);
		}
		printf("\n");

		printf("7-9:\t%5.2f(%5.2f | %5.2f)\t",
				36.0*summ[NORM].partial[bonus-MIN_BONUS]/COUNT_D,
					36.0*(summ[NORM].partial[bonus-MIN_BONUS-1] - summ[NORM].partial[bonus-MIN_BONUS])/COUNT_D,
					36.0*(summ[NORM].partial[bonus-MIN_BONUS+1] - summ[NORM].partial[bonus-MIN_BONUS])/COUNT_D);
		for (j = 0; j < USE_CASE_COUNT; j++) {
			if (j != NORM)
				printf("%5.2f(%5.2f)\t",
					36.0*summ[j].partial[bonus-MIN_BONUS]/COUNT_D,
					36.0*(summ[j].partial[bonus-MIN_BONUS] - summ[NORM].partial[bonus-MIN_BONUS])/COUNT_D);
		}
		printf("\n"); 

		printf("10+:\t%5.2f(%5.2f | %5.2f)\t",
				36.0*summ[NORM].succ[bonus-MIN_BONUS]/COUNT_D,
					36.0*(summ[NORM].succ[bonus-MIN_BONUS-1] - summ[NORM].succ[bonus-MIN_BONUS])/COUNT_D,
					36.0*(summ[NORM].succ[bonus-MIN_BONUS+1] - summ[NORM].succ[bonus-MIN_BONUS])/COUNT_D);
		for (j = 0; j < USE_CASE_COUNT; j++) {
			if (j != NORM)
				printf("%5.2f(%5.2f)\t",
					36.0*summ[j].succ[bonus-MIN_BONUS]/COUNT_D,
					36.0*(summ[j].succ[bonus-MIN_BONUS] - summ[NORM].succ[bonus-MIN_BONUS])/COUNT_D);
		}
		printf("\n");  

		printf("sum:\t%5.2f(%5.2f | %5.2f)\t",
				36.0*summ[NORM].res[bonus-MIN_BONUS]/COUNT_D,
					36.0*(summ[NORM].res[bonus-MIN_BONUS-1] - summ[NORM].res[bonus-MIN_BONUS])/COUNT_D,
					36.0*(summ[NORM].res[bonus-MIN_BONUS+1] - summ[NORM].res[bonus-MIN_BONUS])/COUNT_D);
		for (j = 0; j < USE_CASE_COUNT; j++) {
			if (j != NORM)
				printf("%5.2f(%5.2f)\t",
					36.0*summ[j].res[bonus-MIN_BONUS]/COUNT_D,
					36.0*(summ[j].res[bonus-MIN_BONUS] - summ[NORM].res[bonus-MIN_BONUS])/COUNT_D);
		}
		printf("\n");  
	}

	return 0;
}



